//! Frozen, serialisable state
use super::*;
use cryptohelpers::sha256::{Sha256Hash, self};

//use futures::prelude::*;
use std::io;
use tokio::prelude::*;

/// An immutable image of `State`.
//TODO: Implement this when `State` is solidified and working
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
struct FreezeInner
{
    users: HashSet<User>,
    posts: HashSet<(UserID, Post)>,
}

const FREEZE_CHK: &[u8; 4] = b"REI\0";

/// Metadata derived from `FreezeInner`'s CBOR serialisation.
///
/// This is written and read as-is.
#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord, Copy)]
struct FreezeMetadata
{
    chk: [u8; 4],
    version: u32, //Version, //TODO: Convert env!(version) into type
    body_size: u64,
    compressed: bool, //TODO: Should body be compressed? If so, with what?
    body_hash: Sha256Hash,
}

impl FreezeMetadata
{
    #[inline] pub fn len(&self) -> usize
    {
	self.body_size.try_into().expect("Length exceeded limit of `usize`")
    }
    #[inline] fn new(from: &[u8]) -> Self
    {
	Self {
	    chk: *FREEZE_CHK,
	    version: defaults::VERSION,
	    body_size: from.len().try_into().unwrap(), //this should never fail, right?
	    compressed: false,
	    body_hash: sha256::compute_slice(from),
	}
    }
    /// Write this metadata to an async stream, return the number of bytes written.
    pub async fn write_to(&self, mut to: impl tokio::io::AsyncWrite + Unpin) -> io::Result<usize>
    {
	macro_rules! write_all
	{
	    ($expr:expr) => {
		{
		    let bytes = $expr;
		    let bytes = &bytes[..];
		    to.write_all(bytes).await?;
		    bytes.len()
		}
	    }
	}
	let done =
	    write_all!(self.chk) +
	    write_all!(u32::to_le_bytes(self.version)) +
	    write_all!(u64::to_le_bytes(self.body_size)) +
	    write_all!(if self.compressed {[1]} else {[0]}) +
	    write_all!(self.body_hash.as_ref());

	Ok(done)
    }
    /// Read a metadata object from an aynsc stream, without verifying any of its fields
    async fn read_from_unchecked(mut from: impl tokio::io::AsyncRead + Unpin) -> io::Result<Self>
    {
	macro_rules! read_exact
	{
	    ($num:expr) => {
		{
		    let mut buf = [0u8; $num];
		    from.read_exact(&mut buf[..]).await?;
		    buf
		}
	    };
	    (type $type:ty) => {
		{
		    read_exact!(std::mem::size_of::<$type>())
		}
	    };
	}
	Ok(
	    Self {
		chk: read_exact!(4),
		version: u32::from_le_bytes(read_exact!(type u32)),
		body_size: u64::from_le_bytes(read_exact!(type u64)),
		compressed: if read_exact!(1)[0] == 0 { false } else { true },
		body_hash: Sha256Hash::from(read_exact!(type Sha256Hash)),
	    }
	)
    }
    /// Read a metadata object from an async stream, verifying its fields.
    ///
    /// # Note
    /// The `body_hash` field must be verified *after* reading the body.
    pub async fn read_from(from: impl tokio::io::AsyncRead + Unpin) -> eyre::Result<Self>
    {
	macro_rules! enforce {
	    ($expr:expr; $err:expr) => {
		{
		    if ! $expr {
			return Err($err)
		    }
		}
	    }
	}
	let this = Self::read_from_unchecked(from).await
	    .wrap_err(eyre!("Failed to read data from stream"))?;

	macro_rules! enforce {
	    ($expr:expr, $err:expr) => {
		{
		    if ! $expr {
			Err(eyre!($err))
			    .with_section(|| format!("{:?}", this).header("Metadata was"))
		    } else {
			Ok(())
		    }
		}
	    }
	}

	enforce!(&this.chk == FREEZE_CHK, "Check value was invalid")
	    .with_section(|| format!("{:?}", FREEZE_CHK).header("Expected"))
	    .with_section(|| format!("{:?}", &this.chk).header("Got"))
	    .with_suggestion(|| "Was this the correct type of file you wanted to load?")?;
	
	enforce!(this.body_size <= defaults::MAX_IMAGE_READ_SIZE as u64, "Body size exceeded max")
	    .with_section(|| format!("{}", &this.body_size).header("Size read was"))
	    .with_section(|| format!("{}", defaults::MAX_IMAGE_READ_SIZE).header("Max size allowed is"))
	    .with_warning(|| "This may indicate file corruption")?;
	
	enforce!(this.version <= defaults::VERSION, "Unsupported version")
	    .with_section(|| this.version.to_string().header("Read version was"))
	    .with_section(|| defaults::VERSION.to_string().header("Current version is"))
	    .with_suggestion(|| "This file may have been created with a newer version of the program. Try updating the program.")?;
	
	
	Ok(this)
    }

    /// Verify the hash of this metadata by computing the hash of `from` and checking.
    ///
    /// # Notes
    /// It is recommended to to this within a tokio `spawn_blocking` or `block_in_place` closure, as the hashing operation may take a while.
    pub fn verify_hash_blocking(&self, from: impl AsRef<[u8]>) -> bool
    {
	sha256::compute_slice(from) == self.body_hash
    }
}

/// An immutable `State` image
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Freeze
{
    //metadata: FreezeMetadata, //written as-is, calculated from body
    //body: Vec<u8>,  // `FreezeInner` written with CBOR
    inner: FreezeInner, // Dumped to Vec<u8> as CBOR, then FreezeMetadata is calculated from this binary data. Then metadata is written, then the binary blob is written.
}

impl Freeze
{
    /// Generate the output to write.
    ///
    /// This might take a while, and is done synchronously. Recommended to run on blocking task.
    fn gen_output(&self) -> serde_cbor::Result<(FreezeMetadata, Vec<u8>)>
    {
	let body = serde_cbor::to_vec(&self.inner)?;
	Ok((FreezeMetadata::new(&body[..]), body))
    }

    /// Read `Freeze` from an input stream
    pub async fn read_async(mut input: impl tokio::io::AsyncRead + Unpin) -> eyre::Result<Self>
    {
	let inner: FreezeInner = {
	    let unchecked_meta = FreezeMetadata::read_from(&mut input)
		.await
		.wrap_err(eyre!("Failed to read metadata from stream"))?;
	    let mut body: Vec<u8> = vec![0; unchecked_meta.len()];
	    input.read_exact(&mut body[..]).await
		.wrap_err(eyre!("Failed to read body from stream"))
		.with_section(|| format!("{:?}", unchecked_meta).header("Metadata was"))?;
	    
	    tokio::task::spawn_blocking(move || {
		if !unchecked_meta.verify_hash_blocking(&body) {
		    Err(eyre!("Failed to verify metadata hash"))
		} else {
		    serde_cbor::from_slice(&body)
			.wrap_err(eyre!("Failed to deserialise body"))
		}
	    }).await
		.wrap_err("Background task panic")?
		.with_section(|| format!("{:?}", unchecked_meta).header("Metadata was"))?
	};
	Ok(Self {
	    inner,
	})
    }

    /// Write `Freeze` into this output stream on the current task.
    ///
    /// This function runs the serialisation and hashing on a the current task, which is synchronous. Recommended to use `into_write_async` instead.
    pub async fn write_async(&self, mut output: impl tokio::io::AsyncWrite + Unpin) -> eyre::Result<()>
    {
	let (meta, body) = self.gen_output()
	    .wrap_err(eyre!("Failed to generate write body"))?;
	meta.write_to(&mut output).await
	    .wrap_err(eyre!("Failed to write metadata to output stream"))
	    .with_section(|| format!("{:?}", meta).header("Metadata was"))?;
	output.write_all(&body[..]).await
	    .wrap_err(eyre!("Failed to write whole body to output stream"))?;
	Ok(())
    }
    
    /// Consume this `Freeze` into this output stream.
    ///
    /// This function runs the serialisation and hashing on a background blocking task instead of on the current one.
    pub async fn into_write_async(self, mut output: impl tokio::io::AsyncWrite + Unpin) -> eyre::Result<()>
    {
	let (meta, body) = tokio::task::spawn_blocking(move || self.gen_output())
	    .await
	    .wrap_err(eyre!("Background task panic"))?
	    .wrap_err(eyre!("Failed to generate write body"))?;
	meta.write_to(&mut output).await
	    .wrap_err(eyre!("Failed to write metadata to output stream"))
	    .with_section(|| format!("{:?}", meta).header("Metadata was"))?;
	output.write_all(&body[..]).await
	    .wrap_err(eyre!("Failed to write whole body to output stream"))?;
	Ok(())
    }
}

impl State
{
    /// Create an image from the state.
    ///
    /// # Locks
    /// This method holds the read lock of Oneesan, it also holds the read lock of all posts and users.
    /// This will prevent any writes while the `Freeze` is being created, and will also yield the current task until all write operations on `State` are completed.
    ///
    /// # Panics
    /// If the internal state is incorrect.
    pub async fn freeze(&self) -> Freeze
    {
	let onee = self.0.oneesan.read().await;
	use std::ops::Deref;

	// this might be kinda expensive. should we offload this?
	let post_owner_reverse_lookup: HashMap<Index, UserID> = onee.posts_user_map.iter()
	    .map(|(&y,x)| x.iter().map(move |idx| (*idx,y)))
	    .flatten()
	    .collect();
	let (posts, users) = tokio::join!(
	    stream::iter(onee.posts.iter()).then(|(post_idx, shared)| {
		let owner_id = *post_owner_reverse_lookup.get(&post_idx).unwrap();
		async move { (owner_id, Post::clone(shared.read().await.deref())) }
	    }).collect(),
	    stream::iter(onee.users.iter()).then(|(_, shared)| async move { User::clone(shared.read().await.deref()) }).collect()
	);
	let inner = FreezeInner {
	    posts, users
	};
	Freeze {
	    inner,
	}
    }
}

impl Freeze
{
    /// Create a working `State` from this image.
    ///
    /// This clones all posts and users in the image. Use `into_state` to move into a state.
    pub fn unfreeze(&self) -> State
    {
	let mut users = Arena::with_capacity(self.inner.users.len());
	let mut posts = Arena::with_capacity(self.inner.posts.len());
	let mut posts_map = HashMap::with_capacity(self.inner.posts.len());
	let mut users_map = HashMap::with_capacity(self.inner.users.len());
	let mut posts_user_map = HashMap::with_capacity(self.inner.users.len());
	
	for (owner_id, post) in self.inner.posts.iter()
	{
	    let idx = posts.insert(Arc::new(RwLock::new(post.clone())));
	    posts_user_map.entry(*owner_id).or_insert_with(|| MaybeVec::new()).push(idx);
	    posts_map.insert(*post.post_id(), idx);
	}

	for user in self.inner.users.iter()
	{
	    let idx = users.insert(Arc::new(RwLock::new(user.clone())));
	    users_map.insert(*user.id(), idx);
	}
	
	State(Arc::new(Inner {
	    oneesan: RwLock::new(Oneesan {
		users,
		posts,
		posts_map,
		users_map,
		posts_user_map
	    })
	}))
    }
    
    /// Convert this image into a new `State`.
    pub fn into_state(self) -> State
    {
	let mut users = Arena::with_capacity(self.inner.users.len());
	let mut posts = Arena::with_capacity(self.inner.posts.len());
	let mut posts_map = HashMap::with_capacity(self.inner.posts.len());
	let mut users_map = HashMap::with_capacity(self.inner.users.len());
	let mut posts_user_map = HashMap::with_capacity(self.inner.users.len());
	
	for (owner_id, post) in self.inner.posts.into_iter()
	{
	    let post_id = *post.post_id();
	    let idx = posts.insert(Arc::new(RwLock::new(post)));
	    posts_user_map.entry(owner_id).or_insert_with(|| MaybeVec::new()).push(idx);
	    posts_map.insert(post_id, idx);
	}

	for user in self.inner.users.into_iter()
	{
	    let user_id = *user.id();
	    let idx = users.insert(Arc::new(RwLock::new(user)));
	    users_map.insert(user_id, idx);
	}
	
	State(Arc::new(Inner {
	    oneesan: RwLock::new(Oneesan {
		users,
		posts,
		posts_map,
		users_map,
		posts_user_map
	    })
	}))
    }
}

impl From<Freeze> for State
{
    #[inline] fn from(from: Freeze) -> Self
    {
	from.into_state()
    }
}

#[cfg(test)]
mod tests
{
    #[tokio::test]
    async fn read_write()
    {
	let mut buffer = Vec::new();
	let state = super::State::new();
	let freeze = state.freeze().await;

	freeze.write_async(&mut buffer).await.unwrap();
	let freeze2 = super::Freeze::read_async(&mut &buffer[..]).await.unwrap();
	assert_eq!(freeze, freeze2);

	let mut buffer2 = Vec::new();
	freeze2.into_write_async(&mut buffer2).await.unwrap();
	assert_eq!(buffer, buffer2);
	
    }
    #[tokio::test]
    async fn creation()
    {
	let state = super::State::new();
	let freeze = state.freeze().await;
	let state2 = freeze.unfreeze();
	let freeze2 = state2.freeze().await;
	assert_eq!(freeze, freeze2);
	let _state3 = freeze2.into_state();
    }
    
}
